// Generated by CoffeeScript 2.4.1
(function() {
  var $, Child, Handlebars, Health, Helpers, Http, Opts, Os, Process, Server, Shell, Worker, log, verbose;

  $ = require('bling');

  Os = require("os");

  Shell = require('shelljs');

  Handlebars = require("handlebars");

  Health = require('./health');

  Process = require('./process');

  Helpers = require('./helpers');

  Http = require('./http');

  Opts = require('./opts');

  log = $.logger("[child]");

  verbose = function() {
    var err;
    try {
      if (Opts.verbose) {
        return log.apply(null, arguments);
      }
    } catch (error) {
      err = error;
      return log("verbose error:", $.debugStack(err));
    }
  };

  Child = (function() {
    var toString;

    class Child {
      constructor(opts, index) {
        $.extend(this, {
          opts: opts,
          index: index,
          process: null,
          started: $.extend($.Promise(), {
            attempts: 0,
            timeout: null
          })
        });
        this.log = $.logger(this.toString());
        this.log.verbose = () => {
          var err;
          try {
            if (Opts.verbose) {
              return this.log.apply(null, arguments);
            }
          } catch (error) {
            err = error;
            return this.log("verbose error:", $.debugStack(err));
          }
        };
        this.started.then(() => {
          return this.last_start = $.now;
        });
      }

      uptime() {
        if (this.last_start) {
          return $.now - this.last_start;
        } else {
          return 0;
        }
      }

      uptimeString() {
        var hours, minutes, seconds;
        return $(hours = this.uptime() / 3600000, minutes = (hours % 1) * 60, seconds = (minutes % 1) * 60).map(function() {
          return $.padLeft(String(Math.floor(this)), 2, "0");
        }).join(":");
      }

      start() {
        var cmd, fail, on_data;
        try {
          return this.started;
        } finally {
          fail = (msg) => {
            return this.started.reject(msg);
          };
          if (++this.started.attempts > this.opts.restart.maxAttempts) {
            fail("too many attempts");
          } else {
            // The @started.timeout keeps track of how long we have been waiting for the @started promise to finish
            // For servers, this happens when the PORT is owned.
            // For workers, this happens as soon as the shell command finishes.
            clearTimeout(this.started.timeout);
            this.started.timeout = setTimeout((() => {
              return this.started.attempts = 0;
            }), this.opts.restart.maxInterval);
            // Execute the shell command that will launch the new child.
            log("shell >", cmd = `env ${this.env()} bash -c 'cd ${this.opts.cd} && ${this.opts.command}'`);
            this.process = Shell.exec(cmd, {
              silent: true,
              async: true
            }, $.identity);
            // When the new child process exits, run our @onExit handler
            this.process.on("exit", (err, signal) => {
              return this.onExit(err, signal);
            });
            // All data generated by the new child (via stdout, stderr) is captured and logged
            on_data = (prefix) => {
              return (data) => {
                var i, len, line, ref, results;
                ref = String(data).split(/\n/);
                results = [];
                for (i = 0, len = ref.length; i < len; i++) {
                  line = ref[i];
                  if (line.length) {
                    results.push(this.log(prefix + line));
                  }
                }
                return results;
              };
            };
            this.process.stdout.on("data", on_data(""));
            this.process.stderr.on("data", on_data("(stderr) "));
            // If the Shell.exec above didn't produce a real process, then die.
            if (!this.process.pid) {
              fail("no pid");
            }
          }
        }
      }

      // IMPORTANT NOTE: this function does not resolve @started on it's own,
      // a sub-class like Server or Worker is expected to @started.resolve()
      stop(signal) {
        var err, p;
        try {
          return p = $.Promise();
        } finally {
          this.started.attempts = 2e308;
          this.expectedExit = true;
          if (this.process) {
            try {
              Process.killTree(this.process.pid, signal).then(p.resolve, p.reject);
            } catch (error) {
              err = error;
              log("Error calling killTree:", $.debugStack(err));
            }
          } else {
            p.resolve();
          }
        }
      }

      restart() {
        var p, restart;
        try {
          return p = $.Promise().then(() => {
            return this.last_start = $.now;
          });
        } finally {
          if (this.process == null) {
            log("Starting fresh child (no existing process)");
            this.start().then(p.resolve, p.reject);
          } else {
            restart = () => {
              var err;
              try {
                log("Restarting child...");
                this.process = null;
                this.started.reset();
                this.started.attempts = 0;
                return this.start().then(p.resolve, p.reject);
              } catch (error) {
                err = error;
                log("restart error:", $.debugStack(err));
                return p.reject(err);
              }
            };
            log("Killing existing process", this.process.pid);
            this.expectedExit = true;
            Process.killTree(this.process.pid, "SIGTERM").wait(this.opts.restart.gracePeriod, (err) => {
              try {
                if (String(err) === "Error: timeout") {
                  log(`Child failed to die within ${this.opts.restart.gracePeriod}ms, using SIGKILL`);
                  return Process.killTree(this.process.pid, "SIGKILL").then(restart, p.reject);
                } else if (err) {
                  return p.reject(err);
                } else {
                  return restart();
                }
              } catch (error) {
                err = error;
                log("restart error during kill tree:", $.debugStack(err));
                return p.reject(err);
              }
            });
          }
        }
      }

      onExit(code, signal) {
        var err;
        try {
          signal = $.is('number', code) ? code - 128 : Process.getSignalNumber(signal);
          this.log(`Child exited (signal=${signal})`, this.expectedExit ? "(expected)" : "");
          // add a small restart cooldown
          return $.delay(1000, () => {
            if (!this.expectedExit) {
              this.restart();
            }
            return this.expectedExit = false;
          });
        } catch (error) {
          err = error;
          return this.log("child.onExit error:", $.debugStack(err));
        }
      }

      env() {
        var err, key, val;
        try {
          return ((function() {
            var ref, results;
            ref = this.opts.env;
            results = [];
            for (key in ref) {
              val = ref[key];
              if (val != null) {
                results.push(`${key}="${val}"`);
              }
            }
            return results;
          }).call(this)).join(" ");
        } catch (error) {
          err = error;
          return log("env error:", $.debugStack(err));
        }
      }

      static defaults(opts) {
        opts = $.extend(Object.create(null), {
          cd: ".",
          command: "node index.js",
          count: -1,
          env: {}
        }, opts);
        opts.count = parseInt(opts.count, 10);
        while (opts.count < 0) {
          opts.count += Os.cpus().length;
        }
        opts.count || (opts.count = 1);
        // control what happens at (re)start time
        opts.restart = $.extend(Object.create(null), {
          maxAttempts: 5, // failing five times fast is fatal
          maxInterval: 10000, // in what interval is "fast"?
          gracePeriod: 3000, // how long to wait for a forcibly killed process to die
          timeout: 10000 // how long to wait for a newly launched process to start listening on it's port
        }, opts.restart);
        // defaults for the git configuration
        opts.git = $.extend(Object.create(null), {
          enabled: false,
          cd: ".",
          remote: "origin",
          branch: "master",
          command: "git pull {{remote}} {{branch}} || git merge --abort"
        }, opts.git);
        opts.git.command = Handlebars.compile(opts.git.command);
        opts.git.command.inspect = function(level) {
          return '"' + opts.git.command({
            remote: "{{remote}}",
            branch: "{{branch}}"
          }) + '"';
        };
        return opts;
      }

    };

    Child.prototype.toString = toString = function() {
      var err;
      try {
        return `${this.constructor.name}[${this.index}]`;
      } catch (error) {
        err = error;
        return log("toString error:", $.debugStack(err));
      }
    };

    Child.prototype.inspect = toString;

    return Child;

  }).call(this);

  Worker = (function() {
    var workers;

    class Worker extends Child {
      constructor(opts, index) {
        super(opts = Worker.defaults(opts), index);
        workers.push(this);
        this.log = opts.prefix ? $.logger(`${this.opts.prefix}[${this.index}]`) : $.logger(this.toString());
      }

      start() {
        super.start();
        return this.started.resolve();
      }

    };

    Http.get("/workers", function(req, res) {
      var worker;
      return res.pass(`[${((function() {
        var i, len, ref, ref1, results;
        results = [];
        for (i = 0, len = workers.length; i < len; i++) {
          worker = workers[i];
          results.push(`[${(ref = (ref1 = worker.process) != null ? ref1.pid : void 0) != null ? ref : "DEAD"}, ${worker.port}]`);
        }
        return results;
      })()).join(",\n")}]`);
    });

    Http.get("/workers/restart", function(req, res) {
      var i, len, worker;
      for (i = 0, len = workers.length; i < len; i++) {
        worker = workers[i];
        worker.restart();
      }
      return res.redirect(302, "/console#restarting-workers");
    });

    workers = [];

    Worker.defaults = Child.defaults;

    return Worker;

  }).call(this);

  Server = (function() {
    var servers;

    class Server extends Child {
      constructor(opts, index) {
        var name;
        super(opts = Server.defaults(opts), index);
        this.port = opts.port + index;
        this.log = $.logger(`(${this.opts.prefix || this.opts.cd}:${this.port})`);
        (servers[name = opts.port] != null ? servers[name] : servers[name] = []).push(this);
      }

      // wrap the default start function
      start() {
        this.started.then(() => {
          return Health.monitor(this.port, this.process.pid, this.opts.check);
        });
        try {
          return this.started;
        } finally {
          // find any process that is listening on our port
          Process.clearCache().findOne({
            ports: this.port
          }).then((owner) => {
            if (owner != null) {
              this.log("Killing previous owner of", this.port, "PID:", owner.pid);
              return Process.killTree(owner, "SIGKILL").then(() => {
                return this.start(); // port is available, so really start
              });
            } else {
              super.start(); // do the base Child start
              if (!this.process) {
                return this.started.reject("no process");
              } else {
                verbose("Waiting for port", this.port, "to be owned by", this.process.pid);
                if (this.port_poller != null) {
                  this.port_poller.cancel();
                }
                return this.port_poller = Helpers.portIsOwned(this.process.pid, this.port, this.opts.restart.timeout).then(((result) => {
                  if (!($.is('string', result) && ~(result.indexOf('cancelled')))) {
                    verbose(`Port ${this.port} is successfully owned.`);
                    return this.started.resolve();
                  } else {
                    return verbose("Port polling:", result);
                  }
                }), this.started.reject);
              } // if the port is being listened on
            }
          });
        }
      }

      env() {
        return super.env() + `${this.opts.portVariable}="${this.port}"`;
      }

      static defaults(opts) {
        opts = $.extend({
          port: 8001,
          portVariable: "PORT",
          poolName: "shepherd_pool"
        }, Child.defaults(opts));
        opts.port = parseInt(opts.port, 10);
        opts.check = $.extend({
          enabled: false,
          url: "/",
          status: 200,
          contains: null,
          timeout: 3000,
          interval: 10000
        }, opts.check);
        return opts;
      }

    };

    Http.get("/servers", function(req, res) {
      var port, ret, s, v;
      ret = "[";
      for (port in servers) {
        v = servers[port];
        ret += ((function() {
          var i, len, ref, ref1, results;
          results = [];
          for (i = 0, len = v.length; i < len; i++) {
            s = v[i];
            results.push(`[${(ref = (ref1 = s.process) != null ? ref1.pid : void 0) != null ? ref : "DEAD"}, ${s.port}]`);
          }
          return results;
        })()).join(",\n");
      }
      return res.pass(ret + "]");
    });

    Http.get("/servers/restart", function(req, res) {
      $.valuesOf(servers).flatten().select('restart').call();
      return res.redirect(302, "/console#restarting-servers");
    });

    // a map of base port to all Server instances based on that port
    servers = {};

    return Server;

  }).call(this);

  $.extend(module.exports, {Server, Worker});

}).call(this);
