// Generated by CoffeeScript 2.4.1
(function() {
  var $, Opts, Process, http, int, log, monitor, monitors, unmonitor, verbose;

  $ = require('bling');

  http = require('http');

  Opts = require('./opts');

  Process = require('./process');

  log = $.logger("[monitor]");

  verbose = function(...a) {
    if (Opts.verbose) {
      return log(...a);
    }
  };

  // keep a mapping from port to interval (i.e. setInterval) of the monitor
  monitors = Object.create(null);

  int = function(n) {
    return parseInt(n, 10);
  };

  monitor = function(port, pid, check) {
    var fail;
    if (port in monitors) {
      verbose(`IGNORED: attempt to monitor the same port twice (port: ${port})`);
      return;
    }
    // 'check' is the opts.check block from the servers section of the shepherd.json file
    // example:
    // "enabled": true
    // "url": "/health-check"
    // "status": 200
    // "contains": "IM OK"
    // "timeout": 1000
    // "interval": 5000
    if (!check.enabled) {
      return;
    }
    log("Starting health checks for pid:", pid, "port:", port);
    check.interval = int(check.interval);
    check.status = int(check.status);
    check.timeout = int(check.timeout);
    try {
      check.contains = new RegExp(check.contains);
    } catch (error) {}
    fail = function(err) {
      return function(_err) {
        if (port in monitors) {
          log(`Health check failed (port=${port} pid=${pid}), due to`, err, "caused by", _err);
          unmonitor(port);
          return Process.killTree(pid, 15);
        }
      };
    };
    return monitors[port] = $.interval(check.interval, function() {
      var req;
      req = http.request({
        port: port,
        path: check.url
      }, function(res) {
        var buffer;
        if (res.statusCode !== check.status) {
          return fail(`bad status: ${res.statusCode}`)();
        }
        res.setEncoding('utf8');
        if (check.contains != null) {
          buffer = "";
          res.on('data', function(chunk) {
            return buffer += chunk;
          });
          res.on('end', function() {
            if (!check.contains.test(buffer)) {
              return fail(`response body does match regex: ${check.contains.toString()}`)();
            }
          });
          return res.on('error', fail('response error'));
        }
      });
      req.setTimeout(check.timeout);
      req.on('timeout', fail('timeout'));
      req.on('error', fail('request error'));
      return req.end();
    });
  };

  unmonitor = function(port) {
    if (!(port in monitors)) {
      return;
    }
    monitors[port].cancel();
    return delete monitors[port];
  };

  $.extend(module.exports, {monitor, unmonitor});

}).call(this);
