// Generated by CoffeeScript 2.4.1
(function() {
  var $, Convert, Fs, Handlebars, Helpers, Herd, Http, Jade, Opts, Os, Process, Rabbit, Server, Shell, Worker, clean_exit_code, cpuCount, default_stop_timeout, dirty_exit_code, listen_retry_interval, log, renderView, stop_delay, verbose, viewCache, view_cache_timeout;

  [$, Os, Fs, Handlebars, Shell, Process, {Server, Worker}, Http, Opts, Helpers, Rabbit, Convert, Jade] = ['bling', 'os', 'fs', 'handlebars', 'shelljs', './process', './child', './http', './opts', './helpers', 'rabbit-pubsub', './convert', 'jade'].map(require);

  cpuCount = Os.cpus().length;

  // loggers
  log = $.logger(`[herd-${$.random.string(4)}]`);

  verbose = function() {
    if (Opts.verbose) {
      return log.apply(null, arguments);
    }
  };

  // constants
  listen_retry_interval = 200; // ms

  stop_delay = 300; // ms

  default_stop_timeout = Convert(30).seconds.to.ms;

  clean_exit_code = 0;

  dirty_exit_code = 1;

  view_cache_timeout = 3000;

  if ($.config("NODE_ENV") === "production") {
    view_cache_timeout = 60000;
  }

  viewCache = new $.Cache(2e308, view_cache_timeout);

  renderView = function(view, args) {
    view = __dirname + "/../views/" + view;
    if (viewCache.has(view)) {
      return viewCache.get(view)(args);
    } else {
      return viewCache.set(view, Jade.compile(Fs.readFileSync(view)))(args);
    }
  };

  module.exports = Herd = (function() {
    var buildNginxConfig, connectSignals, listen, writeNginxConfig;

    class Herd {
      constructor(opts) {
        var i, index, j, k, l, len, len1, my, r, ref, ref1, ref2, ref3;
        this.opts = Herd.defaults(opts);
        this.shepherdId = [Os.hostname(), this.opts.admin.port].join(":");
        this.children = [];
        connectSignals(this); // register our process signal handlers
        ref = this.opts.servers;
        // create Server objects
        for (i = 0, len = ref.length; i < len; i++) {
          opts = ref[i];
          while (opts.count < 0) {
            opts.count += cpuCount;
          }
          for (index = j = 0, ref1 = opts.count; j < ref1; index = j += 1) {
            verbose("Creating server:", opts, index);
            this.children.push(new Server(opts, index));
          }
        }
        ref2 = this.opts.workers;
        // create Worker objects
        for (k = 0, len1 = ref2.length; k < len1; k++) {
          opts = ref2[k];
          while (opts.count < 0) {
            opts.count += cpuCount;
          }
          for (index = l = 0, ref3 = opts.count; l < ref3; index = l += 1) {
            verbose("Creating worker:", opts, index);
            this.children.push(new Worker(opts, index));
          }
        }
        Http.get("/", function(req, res) {
          var packageFile;
          packageFile = __dirname + "/../package.json";
          return Helpers.readJson(packageFile).wait(function(err, data) {
            if (err) {
              return res.fail(err);
            } else {
              return res.pass({
                name: data.name,
                version: data.version
              });
            }
          });
        });
        Http.get("/tree", function(req, res) {
          return Process.findTree({
            pid: process.pid
          }).then(function(tree) {
            return res.pass(Process.printTree(tree));
          });
        });
        Http.get("/tree/json", function(req, res) {
          return Process.findTree({
            pid: process.pid
          }).then(res.pass);
        });
        Http.get("/console", (req, res) => {
          var start;
          start = $.now;
          return Process.findTree({
            pid: process.pid
          }).then((tree) => {
            var child, len2, m, ref4, uptimes, visit;
            log("reading the process tree took:", $.now - start, "ms");
            uptimes = Object.create(null);
            ref4 = this.children;
            for (m = 0, len2 = ref4.length; m < len2; m++) {
              child = ref4[m];
              uptimes[child.process.pid] = child.uptimeString();
            }
            visit = function(node, parent) {
              var len3, n, ref5, ref6;
              node.uptime = (ref5 = uptimes[node.pid]) != null ? ref5 : parent != null ? parent.uptime : void 0;
              ref6 = node.children;
              for (n = 0, len3 = ref6.length; n < len3; n++) {
                child = ref6[n];
                visit(child, node);
              }
              return null;
            };
            visit(tree);
            return $.Promise.collect($(this.opts.servers).concat(this.opts.workers).map(function(proc) {
              return Process.exec(`cd ${proc.cd} && git log -1 --pretty=format:'%d %h - %f - %ar'`).then(function(result) {
                return proc.status = result;
              });
            })).then((status) => {
              var err;
              try {
                return res.html(renderView("console.jade", {
                  hostname: Os.hostname(),
                  opts: JSON.stringify(this.opts),
                  tree: JSON.stringify(tree),
                  status: JSON.stringify(status)
                }));
              } catch (error) {
                err = error;
                return res.pass($.debugStack(err));
              }
            });
          });
        });
        Http.get("/stop", (req, res) => {
          return this.stop("SIGTERM").then((function() {
            res.redirect(302, "/console#stopping");
            return $.delay(stop_delay, process.exit);
          }), res.fail);
        });
        Http.get("/reload", (req, res) => {
          this.restart();
          return res.redirect(302, "/console#reloading");
        });
        Http.get("/reload/:pid", function(req, res) { // restart a single child PID
          return Process.findTree({
            pid: process.pid
          }).then(function(tree) {
            var find_pid, ref4, visit;
            // for safety, search our tree to make sure we are killing a managed PID
            find_pid = parseInt(req.params.pid, 10);
            return (ref4 = tree.children) != null ? ref4.forEach(visit = function(node) {
              var ref5;
              if (node.pid === find_pid) {
                console.log("killing", node.pid);
                return Process.kill(node.pid).then((function() {
                  console.log("done, redirecting...");
                  return res.redirect(302, `/console#killed-${node.pid}`);
                }), res.fail);
              } else {
                return (ref5 = node.children) != null ? ref5.forEach(visit) : void 0;
              }
            }) : void 0;
          });
        });
        my = (o) => {
          return $.extend({
            shep: this.shepherdId
          }, o);
        };
        r = this.opts.rabbitmq;
        if (r.enabled && r.url && r.channel) {
          verbose(`Connecting to ${r.url} channel: ${r.channel}...`);
          Rabbit.connect(r.url).wait(function(err) {
            if (err) {
              return log("Connection error:", err);
            }
          });
          Rabbit.subscribe(r.channel, {
            op: "get-status"
          }, (msg) => {
            log("ack receipt of get-status", msg);
            return Process.findTree({
              pid: process.pid
            }).then((tree) => {
              var reply;
              reply = $.extend({
                shep: this.shepherdId
              }, msg, {
                op: "status-reply",
                tree: tree
              });
              Rabbit.publish(r.channel, reply);
              return log("replying to get-status:", reply);
            });
          });
          Rabbit.subscribe(r.channel, my({
            op: "stop"
          }), (msg) => {
            return this.stop("SIGTERM").then(function() {
              return $.delay(stop_delay, process.exit);
            });
          });
          Rabbit.subscribe(r.channel, my({
            op: "reload"
          }), (msg) => {
            return this.restart();
          });
        }
      }

      setStatus(status1, args) {
        this.status = status1;
        verbose("Changing status to:", this.status);
        if (this.opts.rabbitmq.enabled) {
          return Rabbit.publish(this.opts.rabbitmq.channel, {
            op: "set-status",
            status: this.status,
            args: args,
            shep: this.shepherdId
          });
        }
      }

      start(p = $.Promise()) {
        this.setStatus("starting");
        try {
          return p;
        } finally {
          listen(this).then((() => { // start the admin server
            log("Admin server listening on port:", this.opts.admin.port);
            // TODO: respect an admin.enabled=false configuration
            p.then((() => {
              return this.setStatus("started");
            }), ((err) => {
              return this.setStatus(`failed: ${String(err)}`);
            }));
            return writeNginxConfig(this).then(((msg) => { // write the dynamic configuration
              verbose(msg);
              return this.restart().then(p.resolve, p.reject);
            }), p.reject);
          }), p.reject);
        }
      }

      stop(signal, timeout = default_stop_timeout) {
        var child, err, holder, i, len, p, ref;
        this.setStatus("stopping");
        log("Stopping all children with", signal);
        try {
          return p = $.Progress(1).on('progress', function(cur, max) {
            return log(`Stopping progress: ${cur}/${max}`);
          });
        } finally {
          ref = this.children;
          for (i = 0, len = ref.length; i < len; i++) {
            child = ref[i];
            if (!child.process) {
              continue;
            }
            verbose("Attempting to stop child:", child.process.pid, signal);
            try {
              p.include(child.stop(signal));
            } catch (error) {
              err = error;
              log("Error stopping child:", $.debugStack(err));
              p.reject(err);
            }
          }
          holder = setTimeout((function() {
            log(`Failed to stop children within ${timeout} seconds.`);
            return p.reject("timeout");
          }), timeout);
          p.finish(1).then((() => {
            this.setStatus("stopped");
            return clearTimeout(holder);
          }), (err) => {
            return this.setStatus("failed to stop", String(err));
          });
        }
      }

      restart(from = 0, done = $.Promise()) { // perform a careful rolling restart
        var child, next;
        if (from === 0) {
          verbose("Rolling restart starting...");
        }
        try {
          return done;
        } finally {
          next = () => {
            return this.restart(from + 1, done);
          };
          child = this.children[from];
          switch (true) {
            // if the from index is past the end
            case from >= this.children.length:
              verbose("Rolling restart finished.");
              done.resolve();
              break;
            // if there is no such server
            case child == null:
              done.reject(`invalid child index: ${from}`);
              break;
            default:
              verbose("Rolling restart:", from, child.restart().then(next, done.reject));
          }
        }
      }

    };

    // use opts.poolName and opts.nginx.template to render the 'upstream' block for nginx
    buildNginxConfig = function(self) {
      var child, i, len, name, pools, ref, s, servers, upstream;
      s = "";
      pools = Object.create(null);
      ref = self.children;
      for (i = 0, len = ref.length; i < len; i++) {
        child = ref[i];
        if ('poolName' in child.opts && 'port' in child) {
          (pools[name = child.opts.poolName] || (pools[name] = [])).push(child);
        }
      }
      for (upstream in pools) {
        servers = pools[upstream];
        s += self.opts.nginx.template({
          upstream,
          servers,
          pre: "",
          post: ""
        });
      }
      verbose("nginx configuration:\n", s);
      return s;
    };

    // write the nginx config to a file
    writeNginxConfig = function(self) {
      var err, fail, nginx, p, pass;
      nginx = self.opts.nginx;
      try {
        return p = $.Promise();
      } finally {
        fail = function(msg, err) {
          return p.reject(msg + $.debugStack(err));
        };
        pass = function(msg) {
          return p.resolve(msg);
        };
        if (!$.is('bool', nginx.enabled)) {
          fail("Invalid nginx configuration: nginx.enabled is not a boolean.");
        } else if (!$.is('string', nginx.config)) {
          fail("Invalid nginx configuration: nginx.config is not a string.");
        } else if (!$.is('string', nginx.reload)) {
          fail("Invalid nginx configuration: nginx.reload is not a string.");
        } else if ((!nginx.enabled) || (!nginx.config)) {
          pass("Nginx integration not enabled.");
        } else {
          try {
            verbose("Writing nginx configuration to file:", nginx.config);
            Fs.writeFile(nginx.config, buildNginxConfig(self), function(err) {
              if (err) {
                return fail("Failed to write nginx configuration file:", err);
              } else {
                return Process.exec(nginx.reload).wait(function(err) {
                  if (err) {
                    return fail(`Failed to reload nginx (exec: ${nginx.reload}):`, err);
                  } else {
                    return pass("Nginx configuration written.");
                  }
                });
              }
            });
          } catch (error) {
            err = error;
            fail("writeNginxConfig exception:", err);
          }
        }
      }
    };

    listen = function(self, p = $.Promise()) {
      var port;
      port = self.opts.admin.port;
      p.then($.identity, function(err) {
        return self.setStatus(`listen failed: ${String(err)}`);
      });
      try {
        return p;
      } finally {
        Process.clearCache().findOne({
          ports: port
        }).wait(function(err, owner) {
          if (owner != null) {
            log(`Killing old listener on port (${port}): ${owner.pid}`);
            return Process.kill(owner.pid, "SIGTERM").wait(function() {
              log(`Will retry after ${listen_retry_interval} ms`);
              return $.delay(listen_retry_interval, function() {
                return listen(self, p);
              });
            });
          } else {
            return Http.listen(port).then(p.resolve, p.reject);
          }
        });
      }
    };

    connectSignals = function(self) {
      var clean_exit, dirty_exit, i, len, ref, sig;
      clean_exit = function() {
        log("Exiting clean...");
        return process.exit(clean_exit_code);
      };
      dirty_exit = function(err) {
        console.error(err);
        return process.exit(dirty_exit_code);
      };
      ref = ["SIGINT", "SIGTERM"];
      // on SIGINT or SIGTERM, kill everything and die
      for (i = 0, len = ref.length; i < len; i++) {
        sig = ref[i];
        (function(sig) {
          return process.on(sig, function() {
            log("Got signal:", sig);
            return self.stop("SIGKILL").then(clean_exit, dirty_exit);
          });
        })(sig);
      }
      // on SIGHUP, just reload all child procs
      process.on("SIGHUP", function() {
        log("Got signal: SIGHUP... reloading.");
        return self.restart();
      });
      return process.on("exit", function(code) {
        return log("shepherd.on 'exit',", code);
      });
    };

    return Herd;

  }).call(this);

  // make sure a herd object has all the default configuration
  Herd.defaults = function(opts) {
    var base;
    opts = $.extend(Object.create(null), {
      servers: [],
      workers: []
    }, opts);
    opts.rabbitmq = $.extend(Object.create(null), {
      enabled: false,
      url: "amqp://localhost:5672",
      channel: "shepherd"
    }, opts.rabbitmq);
    // TODO: only set enabled=true if nginx is actually installed
    opts.nginx = $.extend(Object.create(null), ((function() {
      switch (Os.platform()) {
        case 'darwin':
          return {
            enabled: true,
            config: "/usr/local/etc/nginx/conf.d/shepherd.conf",
            reload: "launchctl stop homebrew.mxcl.nginx && launchctl start homebrew.mxcl.nginx"
          };
        case 'linux':
          return {
            enabled: true,
            config: "/etc/nginx/conf.d/shepherd.conf",
            reload: "/etc/init.d/nginx reload"
          };
        default:
          return {
            enabled: false,
            config: "/dev/null",
            reload: "echo WARN: I don't support nginx on platform: " + Os.platform()
          };
      }
    })()), opts.nginx);
    (base = opts.nginx).template || (base.template = "upstream {{upstream}} {\n	{{pre}}\n	{{#each servers}}\n	server 127.0.0.1:{{this.port}} weight=1;\n	{{/each}}\n	{{post}}\n	keepalive 32;\n}");
    if (!$.is('function', opts.nginx.template)) {
      opts.nginx.template = Handlebars.compile(opts.nginx.template);
      opts.nginx.template.inspect = function(level) { // use a mock rendering as standard output
        return '"' + opts.nginx.template({
          upstream: "{{upstream}}",
          pre: "{{#each servers}}",
          servers: [
            {
              port: "{{this.port}}"
            }
          ],
          post: "{{/each}}"
        }) + '"';
      };
    }
    // the http server listens for REST calls and web hooks
    opts.admin = $.extend(Object.create(null), {
      enabled: true,
      port: 9000
    }, opts.admin);
    verbose("Using configuration:", JSON.stringify(opts, null, '  '));
    return opts;
  };

}).call(this);
