// Generated by CoffeeScript 2.4.1
(function() {
  //!/usr/bin/env node
  var $, Process, Shell, attach_ports, log, lsof_cmd, port, psCache, ps_cmd, ps_parse, signals;

  $ = require('bling');

  Shell = require('shelljs');

  Process = module.exports;

  log = $.logger("[Process]");

  Process.exec = function(cmd, verbose) {
    var append_output, child, err, p, ret;
    try {
      return p = $.Promise();
    } finally {
      try {
        if (verbose) {
          log("shell >", cmd);
        }
        ret = {
          output: ""
        };
        child = Shell.exec(cmd, {
          silent: true,
          async: true
        }, function(exitCode) {
          var err;
          try {
            if (exitCode !== 0) {
              return p.reject(ret.output === "" ? `Non-zero exit code: ${exitCode}` : ret.output);
            } else {
              return p.resolve(ret.output);
            }
          } catch (error) {
            err = error;
            return log("exec: error handling process exit:", $.debugStack(err));
          }
        });
        child.stdout.on("data", append_output = function(data) {
          return ret.output += String(data);
        });
        child.stderr.on("data", append_output);
      } catch (error) {
        err = error;
        log("exec: error in running process:", $.debugStack(err));
      }
    }
  };

  // for caching the output of 'ps' commands
  // mostly to save time in commands like Process.tree
  // where possibly hundreds of Process.find calls are generated.
  psCache = new $.Cache(2, 500);

  ps_cmd = "ps -eo uid,pid,ppid,pcpu,rss,command";

  ps_parse = function(output) {
    var err, keys;
    try {
      output = output.split('\n').map(function(line) { // split into lines
        return line.split(/[ ]+/).slice(1); // split each line on whitespace
      }).slice(0, -1); // discard the last line?
      // turn the 2D array of proc data into a list of process objects
      keys = output[0].map($.slugize); // parse the first line for the field names
      return output.slice(1).map(function(row) { // for each row
        var e, i, j, key, len, ret, val;
        try {
          return ret = {}; // return an object
        } finally {
// attach an output value to each key for this row
          for (i = j = 0, len = keys.length; j < len; i = ++j) {
            key = keys[i];
            if (i === keys.length - 1) { // the last value (the command) is all concatenated together
              ret[key] = row.slice(i).join(' ');
            } else {
              val = row[i];
              try {
                val = parseFloat(val);
                if (!((typeof val === 'number') && isFinite(val))) { // revert the value on a soft parsing (NaN, Infinity, etc)
                  val = row[i];
                }
              } catch (error) {
                e = error;
                val = row[i];
              } finally {
                ret[key] = val;
              }
            }
            if (ret[key] == null) {
              log("ps_parse failed to parse line:", key, i, row[i], ret[key]);
            }
          }
        }
      });
    } catch (error) {
      err = error;
      return log("ps_parse error:", $.debugStack(err));
    }
  };

  // given the output of ps_parse, use "lsof" to attach listening ports
  lsof_cmd = "lsof -Pni | grep LISTEN || exit 0";

  attach_ports = function(procs) {
    var attached, err, index, j, len, proc;
    try {
      return attached = $.Promise();
    } finally {
      try {
        index = Object.create(null);
        for (j = 0, len = procs.length; j < len; j++) {
          proc = procs[j];
          index[proc.pid] = proc;
          proc.ports = [];
        }
        Process.exec(lsof_cmd).wait(function(err, output) {
          var k, len1, line, pid, port, ref;
          if (err) {
            log("exec error:", $.debugStack(err));
            process.exit(1);
          }
          if (!output) {
            return attached.resolve(procs);
          }
          try {
            ref = output.split(/\n/g);
            for (k = 0, len1 = ref.length; k < len1; k++) {
              line = ref[k];
              line = line.split(/\s+/g);
              if (line.length < 8) {
                continue;
              }
              pid = parseInt(line[1], 10);
              port = parseInt(line[8].split(/:/)[1], 10);
              if (!(pid in index)) {
                index[pid] = {
                  pid: pid,
                  ports: []
                };
              }
              try {
                index[pid].ports.push(port);
              } catch (error) {
                err = error;
                log(err, pid, index[pid]);
              }
            }
            return attached.resolve(procs);
          } catch (error) {
            err = error;
            return log("attach_ports error while parsing output:", $.debugStack(err));
          }
        });
      } catch (error) {
        err = error;
        log("attach_ports error:", $.debugStack(err));
      }
    }
  };

  Process.clearCache = function() {
    psCache.del(ps_cmd);
    return Process;
  };

  Process.find = function(query) {
    var err, p;
    try {
      return p = $.Promise();
    } finally {
      try {
        query = (function() {
          switch ($.type(query)) {
            case "string":
              return {
                cmd: new RegExp(query)
              };
            case "number":
              return {
                pid: query
              };
            default:
              return query;
          }
        })();
        if (psCache.has(ps_cmd)) {
          p.resolve(psCache.get(ps_cmd).filter(function(item) {
            return $.matches(query, item);
          }));
        } else {
          Process.exec(ps_cmd).then((function(output) {
            return attach_ports(ps_parse(output)).then((function(procs) {
              var err;
              try {
                return p.resolve(psCache.set(ps_cmd, procs).filter(function(item) {
                  return $.matches(query, item);
                }));
              } catch (error) {
                err = error;
                return log("find error in results:", $.debugStack(err));
              }
            }), p.reject);
          }), p.reject);
        }
      } catch (error) {
        err = error;
        log("find error:", $.debugStack(err));
      }
    }
  };

  Process.findOne = function(query) {
    var p;
    try {
      return p = $.Promise();
    } finally {
      Process.find(query).wait(function(err, out) {
        if (err) {
          return p.reject(err);
        } else {
          try {
            return p.resolve(out[0]);
          } catch (error) {
            err = error;
            return log("findOne error:", $.debugStack(err));
          }
        }
      });
    }
  };

  Process.findTree = function(query) {
    var p;
    try {
      return p = $.Promise();
    } finally {
      Process.findOne(query).then(function(proc) {
        return Process.tree(proc).then(p.resolve, p.reject);
      });
    }
  };

  Process.signals = signals = {
    SIGHUP: 1,
    SIGINT: 2,
    SIGKILL: 9,
    SIGTERM: 15,
    HUP: 1,
    INT: 2,
    KILL: 9,
    TERM: 15
  };

  Process.getSignalNumber = function(signal) {
    var ref;
    return (ref = signals[signal]) != null ? ref : ($.is('number', signal) ? signal : 15);
  };

  Process.kill = function(pid, signal) {
    var err;
    try {
      return Process.exec(`kill -${Process.getSignalNumber(signal)} ${pid}`);
    } catch (error) {
      err = error;
      return log("kill error:", $.debugStack(err));
    }
  };

  Process.tree = function(proc) {
    var p, q;
    try {
      return q = $.Promise();
    } finally {
      p = $.Progress(1);
      if (proc) {
        Process.find({
          ppid: proc.pid
        }).then((function(children) {
          var child, err, j, len;
          try {
            proc.children = children;
            for (j = 0, len = children.length; j < len; j++) {
              child = children[j];
              p.include(Process.tree(child));
            }
            return p.resolve(1, proc);
          } catch (error) {
            err = error;
            return log("tree error:", $.debugStack(err));
          }
        }), q.reject);
      }
      p.then((function() {
        return q.resolve(proc);
      }), q.reject);
    }
  };

  Process.walk = function(node, visit, depth = 0) {
    var child, err, j, len, p, ref;
    try {
      return p = $.Progress(1);
    } finally {
      try {
        p.include(visit(node, depth));
      } catch (error) {
        err = error;
        log("walk error (in visit):", $.debugStack(err));
        p.reject(err);
      }
      ref = node.children;
      for (j = 0, len = ref.length; j < len; j++) {
        child = ref[j];
        p.include(Process.walk(child, visit, depth + 1));
      }
      p.finish(1);
    }
  };

  Process.killTree = function(proc, signal) {
    var err, fail, p, tokill;
    try {
      return p = $.Promise();
    } finally {
      try {
        signal = Process.getSignalNumber(signal);
        proc = (function() {
          switch ($.type(proc)) {
            case 'string':
            case 'number':
              return {
                pid: proc
              };
            default:
              return proc;
          }
        })();
        tokill = [];
        fail = function(msg, err) {
          log(msg, $.debugStack(err));
          return p.reject(err);
        };
        Process.tree(proc).then((function(tree) {
          var err;
          try {
            Process.walk(tree, function(node) {
              if (node.pid) {
                tokill.push(node.pid);
              } else {
                fail("killTree invalid node (no pid):", node);
              }
              return null;
            });
            if (tokill.length) {
              return Process.exec(`kill -${signal} ${tokill.join(' ')} &> /dev/null`).wait(function(err, result) {
                return p.resolve();
              });
            }
          } catch (error) {
            err = error;
            return fail("killTree error while walking:", err);
          }
        }), p.reject);
      } catch (error) {
        err = error;
        fail("killTree error:", err);
      }
    }
  };

  Process.summarize = function(proc) { // currently kind of worthless, needs to use depth
    var p;
    proc.rss = proc.cpu = 0;
    try {
      return p = $.Promise();
    } finally {
      Process.tree(proc).then(function(tree) {
        Process.walk(tree, function(node, depth) {
          proc.rss += node.rss; // sum values upwards
          return proc.cpu += node.cpu;
        });
        return p.resolve(tree);
      });
    }
  };

  Process.printTree = function(proc, indent, spacer) {
    var child, err, j, len, ref, ref1, ret;
    try {
      spacer || (spacer = " \\_");
      indent || (indent = "* ");
      ret = indent + proc.pid + " " + proc.command;
      if ((ref = proc.ports) != null ? ref.length : void 0) {
        ret += " [:" + proc.ports.join(", :") + "]";
      }
      ret += ` {mem: ${$.commaize(proc.rss)}kb cpu: ${proc.cpu}%}\n`;
      indent = spacer + indent;
      ref1 = proc.children;
      for (j = 0, len = ref1.length; j < len; j++) {
        child = ref1[j];
        ret += Process.printTree(child, indent, "   ");
      }
      indent.replace(/^   /, '');
      return ret;
    } catch (error) {
      err = error;
      return log("printTree error:", $.debugStack(err));
    }
  };

  if (require.main === module) {
    port = parseInt(process.argv[2], 10) || 8000;
    log("Tree for owner of:", port);
    Process.find({
      ports: port
    }).then(function(procs) {
      var j, len, proc, results;
      results = [];
      for (j = 0, len = procs.length; j < len; j++) {
        proc = procs[j];
        results.push(Process.tree(proc).then(function(tree) {
          return console.log(Process.printTree(tree));
        }));
      }
      return results;
    });
  }

}).call(this);
