// Generated by CoffeeScript 2.4.1
(function() {
  var $, Fs, Helpers, Herd, Opts, Server, Validate, Worker, d, die, err, exit_soon, log, outputs, reopen, retryCount, verbose;

  Opts = require('./opts');

  if (Opts.daemon) {
    require('daemon')({ // fork into the background
      // stdout: 'inherit', stdin: 'inherit'
      cwd: process.cwd() // work-around indexzero/daemon.node#41, needed for Node 8
    });
  }

  $ = require("bling");

  Fs = require("fs");

  log = $.logger("[shepherd]");

  exit_soon = (n) => {
    return setTimeout((() => {
      return process.exit(n);
    }), 0);
  };

  die = function(...a) {
    log(...a);
    return exit_soon(1);
  };

  verbose = function() {
    if (Opts.verbose) {
      return log.apply(null, arguments);
    }
  };

  // create the default output stream
  outputs = {
    '-': process.stdout
  };

  // create additional output streams requested by the user
  // add addtional { key: WriteStream } pairs here to log to additional places
  if (Opts.O !== "-") {
    try {
      retryCount = 12;
      (reopen = function() {
        if (--retryCount <= 0) {
          console.error("Output stream giving up, too many open failures");
          return;
        }
        return (outputs[Opts.O] = Fs.createWriteStream(Opts.O, {
          flags: 'a',
          mode: 0o640,
          encoding: 'utf8'
        })).on('finish', function(e) {
          console.error("Output stream 'finish' event:", e);
          delete outputs[Opts.O];
          return $.delay(0, reopen);
        }).on('error', function(err) {
          console.error("Output stream 'error' event:", err);
          delete outputs[Opts.O];
          return $.delay(0, reopen);
        });
      })();
    } catch (error) {
      err = error;
      die("Failed to open output stream:", $.debugStack(err));
    }
  }

  $.log.out = function(...a) {
    var data, k, results, s;
    data = a.map($.toString).join(' ') + "\n";
    results = [];
    for (k in outputs) {
      s = outputs[k];
      try {
        results.push(s.write(data, 'utf8'));
      } catch (error) {
        err = error;
        results.push(console.error(`failed to write to output[${k}]:`, $.debugStack(err)));
      }
    }
    return results;
  };

  $.log.enableTimestamps();

  verbose(`Opened output stream to: ${Opts.O}`);

  Helpers = require('./helpers');

  Herd = require('./herd');

  Validate = require('./validate');

  if (Opts.example) {
    d = Herd.defaults();
    ({Server, Worker} = require("./child"));
    d.servers.push(Server.defaults());
    d.workers.push(Worker.defaults());
    console.log(JSON.stringify(d, null, '  '));
    return exit_soon(0);
  }

  if (Opts.P) { // write out a pid file
    verbose("Writing pid file:", Opts.P);
    Fs.writeFileSync(Opts.P, String(process.pid));
  }

  verbose("Reading config file:", Opts.F);

  Helpers.readJson(Opts.F).wait(function(err, config) {
    var errors, ref, ref1, ref2, ref3;
    if (err) {
      die("Failed to open json file:", Opts.F, err.stack);
    }
    config = Herd.defaults(config);
    errors = Validate.isValidConfig(config);
    if (errors.length) {
      die(errors.join("\n"));
    }
    log("Starting new herd, shepherd PID: " + process.pid);
    if ((ref = config.loggly) != null ? ref.enabled : void 0) {
      outputs['loggly'] = require("./loggly").createWriteStream({
        token: config.loggly.token,
        subdomain: config.loggly.subdomain,
        tags: (ref1 = config.loggly.tags) != null ? ref1 : [],
        json: (ref2 = config.loggly.json) != null ? ref2 : false
      });
      verbose(`Opened output stream to Loggly (${config.loggly.subdomain}).`);
    }
    if ((ref3 = config.mongodb) != null ? ref3.enabled : void 0) {
      outputs['mongodb'] = require("./mongodb").createWriteStream({
        url: config.mongodb.url,
        collection: config.mongodb.collection,
        size: config.mongodb.size
      });
      verbose(`Opened output stream to MongoDB (${config.mongodb.url}/${config.mongodb.collection})`);
    }
    return new Herd(config).start().wait(function(err) {
      if (err) {
        return die("Failed to start herd:", $.debugStack(err));
      }
    });
  });

}).call(this);
