// Generated by CoffeeScript 2.4.1
(function() {
  var $, CSON, Fs, JSV, Validate, assert, die, errors, i, isDirectory, isFileCreatable, isFileWritable, j, readableError, ref, schemaFile, testData;

  $ = require('bling');

  Fs = require('fs');

  JSV = require('JSV').JSV.createEnvironment("json-schema-draft-03");

  CSON = require('cson');

  Validate = module.exports;

  schemaFile = function() {
    return $.config.get("SHEPHERD_SCHEMA", __dirname + "/../src/schema.cson");
  };

  readableError = function(err) {
    var field, item, message, path;
    path = $(err.uri.split('#')).last().replace(/^\//, '').split(/\//);
    field = $(path).last();
    item = $(path).slice(0, -1).join(".");
    message = err.message.replace("Instance is ", "");
    if (err.attribute === 'type') {
      message += ` (${err.details[0]})`;
    }
    return `In "${item}", field "${field}" failed validation: ${message}`;
  };

  isFileWritable = function(file) {
    try {
      Fs.accessSync(file, Fs.W_OK);
    } catch (error) {
      return false;
    }
    return true;
  };

  isFileCreatable = function(file) {
    var err, path;
    path = $(file.split('/')).slice(0, -1).join('/');
    try {
      if (!Fs.statSync(path).isDirectory()) {
        return false;
      }
      Fs.accessSync(path, Fs.W_OK);
    } catch (error) {
      err = error;
      return false;
    }
    return true;
  };

  isDirectory = function(path) {
    try {
      return Fs.statSync(path).isDirectory();
    } catch (error) {
      return false;
    }
  };

  Validate.isValidConfig = function(obj) {
    var a, b, errors, i, j, k, l, len, len1, len2, len3, m, ranges, ref, ref1, ref10, ref11, ref2, ref3, ref4, ref5, ref6, ref7, ref8, ref9, schema, server, worker;
    schema = CSON.parseFile(schemaFile());
    errors = JSV.validate(obj, schema).errors.map(readableError);
    if (errors.length) {
      return errors;
    }
    // manual checks:
    // obj.nginx.config points to a writable file
    if (((ref = obj.nginx) != null ? ref.enabled : void 0) && ((ref1 = obj.nginx) != null ? ref1.config : void 0)) {
      if ((!isFileWritable(obj.nginx.config)) && (!isFileCreatable(obj.nginx.config))) {
        return ['In "nginx", field "config" failed validation: file is not writable'];
      }
    }
    // must specify at least one server or one worker
    if ((!obj.servers) && (!obj.workers)) {
      return ['In "", field "servers" failed validation: must specify either "servers" or "workers"'];
    }
    ref3 = (ref2 = obj.servers) != null ? ref2 : [];
    // server and worker 'cd' values point to accessible directories
    for (i = j = 0, len = ref3.length; j < len; i = ++j) {
      server = ref3[i];
      if (server.cd && !isDirectory(server.cd)) {
        return [`In "servers.${i}", field "cd" failed validation: not a directory (${server.cd})`];
      }
    }
    ref5 = (ref4 = obj.workers) != null ? ref4 : [];
    for (i = k = 0, len1 = ref5.length; k < len1; i = ++k) {
      worker = ref5[i];
      if (worker.cd && !isDirectory(worker.cd)) {
        return [`In "workers.${i}", field "cd" failed validation: not a directory (${worker.cd})`];
      }
    }
    // rabbitmq.url should be a parsable url with protocol "amqp"
    if (((ref6 = obj.rabbitmq) != null ? ref6.enabled : void 0) && ((ref7 = obj.rabbitmq) != null ? ref7.url : void 0)) {
      if (((ref8 = $.URL.parse(obj.rabbitmq.url)) != null ? ref8.protocol : void 0) !== "amqp") {
        return ['In "rabbitmq", field "url" failed validation: not a valid amqp:// URL'];
      }
    }
    // server port ranges don't overlap (with each other or with the admin)
    ranges = (function() {
      var l, len2, ref9, results;
      ref9 = obj.servers;
      results = [];
      for (l = 0, len2 = ref9.length; l < len2; l++) {
        server = ref9[l];
        results.push([server.port, server.port + server.count - 1]);
      }
      return results;
    })();
    for (l = 0, len2 = ranges.length; l < len2; l++) {
      a = ranges[l];
      if (obj.admin.enabled && (a[0] <= (ref9 = obj.admin.port) && ref9 <= a[1])) {
        return [`In "servers", port range (${a.join('-')}) conflicts with admin port: ${obj.admin.port}`];
      }
      for (m = 0, len3 = ranges.length; m < len3; m++) {
        b = ranges[m];
        if (a === b) {
          continue;
        }
        if ((a[0] <= (ref10 = b[0]) && ref10 <= a[1]) || (a[0] <= (ref11 = b[1]) && ref11 <= a[1])) {
          return [`In "servers", conflict in port range: ${a.join('-')} overlaps with ${b.join('-')}`];
        }
      }
    }
    return [];
  };

  if (require.main === module) {
    die = function(err) {
      console.error(err);
      return process.exit(1);
    };
    testData = [
      {},
      ['In "", field "servers" failed validation: must specify either "servers" or "workers"'],
      {
        servers: [{}]
      },
      ['In "servers.0", field "command" failed validation: Property is required'],
      {
        servers: [
          {
            command: "echo"
          }
        ]
      },
      [],
      {
        servers: [
          {
            command: "echo"
          }
        ],
        nginx: {
          enabled: "false"
        }
      },
      ['In "nginx", field "enabled" failed validation: not a required type (boolean)'],
      {
        servers: [
          {
            command: "echo"
          }
        ],
        admin: {
          enabled: "false"
        }
      },
      ['In "admin", field "enabled" failed validation: not a required type (boolean)'],
      {
        servers: [
          {
            command: "echo"
          }
        ],
        admin: {
          port: "false"
        }
      },
      ['In "admin", field "port" failed validation: not a required type (number)'],
      {
        servers: [
          {
            command: "echo"
          }
        ],
        rabbitmq: {
          enabled: "false"
        }
      },
      ['In "rabbitmq", field "enabled" failed validation: not a required type (boolean)'],
      {
        servers: [
          {
            command: "echo"
          }
        ],
        rabbitmq: {
          url: null
        }
      },
      ['In "rabbitmq", field "url" failed validation: not a required type (string)'],
      {
        servers: [
          {
            command: "echo"
          }
        ],
        rabbitmq: {
          enabled: true,
          url: "..."
        }
      },
      ['In "rabbitmq", field "url" failed validation: not a valid amqp:// URL'],
      {
        servers: [
          {
            command: "echo"
          }
        ],
        nginx: {
          config: null
        }
      },
      ['In "nginx", field "config" failed validation: not a required type (string)'],
      {
        servers: [
          {
            command: "echo"
          }
        ],
        nginx: {
          enabled: true,
          config: "..."
        }
      },
      ['In "nginx", field "config" failed validation: file is not writable'],
      {
        servers: [
          {
            cd: "no_exist",
            command: "echo"
          }
        ]
      },
      ['In "servers.0", field "cd" failed validation: not a directory (no_exist)'],
      {
        servers: [
          {
            cd: ".",
            command: "echo"
          }
        ]
      },
      []
    ];
    assert = require('assert');
    for (i = j = 0, ref = testData.length - 1; j < ref; i = j += 2) {
      errors = Validate.isValidConfig(testData[i]);
      assert.deepEqual(errors, testData[i + 1]);
    }
  }

}).call(this);
